package org.spongepowered.asm.mixin;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.Sets;
import java.io.File;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import javax.annotation.Resource;
import net.minecraft.launchwrapper.IClassNameTransformer;
import net.minecraft.launchwrapper.IClassTransformer;
import net.minecraft.launchwrapper.ITweaker;
import net.minecraft.launchwrapper.Launch;
import net.minecraft.launchwrapper.LaunchClassLoader;
import org.apache.logging.log4j.Level;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.core.Filter;
import org.apache.logging.log4j.core.Layout;
import org.apache.logging.log4j.core.LogEvent;
import org.apache.logging.log4j.core.Logger;
import org.apache.logging.log4j.core.appender.AbstractAppender;
import org.apache.logging.log4j.core.helpers.Booleans;
import org.spongepowered.asm.launch.Blackboard;
import org.spongepowered.asm.launch.MixinBootstrap;
import org.spongepowered.asm.mixin.Mixins;
import org.spongepowered.asm.mixin.extensibility.IEnvironmentTokenProvider;
import org.spongepowered.asm.mixin.throwables.MixinException;
import org.spongepowered.asm.mixin.transformer.MixinTransformer;
import org.spongepowered.asm.obfuscation.RemapperChain;
import org.spongepowered.asm.util.ITokenProvider;
import org.spongepowered.asm.util.JavaVersion;
import org.spongepowered.asm.util.PrettyPrinter;

public final class MixinEnvironment implements ITokenProvider {

   private static final Set excludeTransformers = Sets.newHashSet(new String[]{"net.minecraftforge.fml.common.asm.transformers.EventSubscriptionTransformer", "cpw.mods.fml.common.asm.transformers.EventSubscriptionTransformer", "net.minecraftforge.fml.common.asm.transformers.TerminalTransformer", "cpw.mods.fml.common.asm.transformers.TerminalTransformer"});
   private static MixinEnvironment currentEnvironment;
   private static MixinEnvironment.Phase currentPhase = MixinEnvironment.Phase.NOT_INITIALISED;
   private static MixinEnvironment.CompatibilityLevel compatibility = (MixinEnvironment.CompatibilityLevel)MixinEnvironment.Option.DEFAULT_COMPATIBILITY_LEVEL.getEnumValue(MixinEnvironment.CompatibilityLevel.JAVA_6);
   private static boolean showHeader = true;
   private static final org.apache.logging.log4j.Logger logger = LogManager.getLogger("mixin");
   private final MixinEnvironment.Phase phase;
   private final String configsKey;
   private final boolean[] options;
   private final Set tokenProviderClasses = new HashSet();
   private final List tokenProviders = new ArrayList();
   private final Map internalTokens = new HashMap();
   private final RemapperChain remappers = new RemapperChain();
   private MixinEnvironment.Side side;
   private List transformers;
   private IClassNameTransformer nameTransformer;
   private String obfuscationContext = null;


   MixinEnvironment(MixinEnvironment.Phase phase) {
      this.phase = phase;
      this.configsKey = "mixin.configs." + this.phase.name.toLowerCase();
      String version = this.getVersion();
      if(version != null && "0.6".equals(version)) {
         if(this.getClass().getClassLoader() != Launch.class.getClassLoader()) {
            throw new MixinException("Attempted to init the mixin environment in the wrong classloader");
         } else {
            this.options = new boolean[MixinEnvironment.Option.values().length];
            MixinEnvironment.Option[] arr$ = MixinEnvironment.Option.values();
            int len$ = arr$.length;

            for(int i$ = 0; i$ < len$; ++i$) {
               MixinEnvironment.Option option = arr$[i$];
               this.options[option.ordinal()] = option.getBooleanValue();
            }

            if(showHeader) {
               showHeader = false;
               this.printHeader(version);
            }

         }
      } else {
         throw new MixinException("Environment conflict, mismatched versions or you didn\'t call MixinBootstrap.init()");
      }
   }

   private void printHeader(Object version) {
      MixinEnvironment.Side side = this.getSide();
      String codeSource = this.getCodeSource();
      logger.info("SpongePowered MIXIN Subsystem Version={} Source={} Env={}", new Object[]{version, codeSource, side});
      if(this.getOption(MixinEnvironment.Option.DEBUG_VERBOSE)) {
         PrettyPrinter printer = new PrettyPrinter(32);
         printer.add("SpongePowered MIXIN (Verbose debugging enabled)").centre().hr();
         printer.kv("Code source", codeSource);
         printer.kv("Internal Version", version);
         printer.kv("Java 8 Supported", Boolean.valueOf(MixinEnvironment.CompatibilityLevel.JAVA_8.isSupported())).hr();
         MixinEnvironment.Option[] arr$ = MixinEnvironment.Option.values();
         int len$ = arr$.length;

         for(int i$ = 0; i$ < len$; ++i$) {
            MixinEnvironment.Option option = arr$[i$];
            StringBuilder indent = new StringBuilder();

            for(int i = 0; i < option.depth; ++i) {
               indent.append("- ");
            }

            printer.kv(option.property, "%s<%s>", new Object[]{indent, option});
         }

         printer.hr().kv("Detected Side", side);
         printer.print(System.err);
      }

   }

   private String getCodeSource() {
      try {
         return this.getClass().getProtectionDomain().getCodeSource().getLocation().toString();
      } catch (Throwable var2) {
         return "Unknown";
      }
   }

   public MixinEnvironment.Phase getPhase() {
      return this.phase;
   }

   /** @deprecated */
   @Deprecated
   public List getMixinConfigs() {
      Object mixinConfigs = (List)Blackboard.get(this.configsKey);
      if(mixinConfigs == null) {
         mixinConfigs = new ArrayList();
         Launch.blackboard.put(this.configsKey, mixinConfigs);
      }

      return (List)mixinConfigs;
   }

   /** @deprecated */
   @Deprecated
   public MixinEnvironment addConfiguration(String config) {
      logger.warn("MixinEnvironment::addConfiguration is deprecated and will be removed. Use Mixins::addConfiguration instead!");
      Mixins.addConfiguration(config, this);
      return this;
   }

   void registerConfig(String config) {
      List configs = this.getMixinConfigs();
      if(!configs.contains(config)) {
         configs.add(config);
      }

   }

   /** @deprecated */
   @Deprecated
   public MixinEnvironment registerErrorHandlerClass(String handlerName) {
      Mixins.registerErrorHandlerClass(handlerName);
      return this;
   }

   public MixinEnvironment registerTokenProviderClass(String providerName) {
      if(!this.tokenProviderClasses.contains(providerName)) {
         try {
            Class th = Class.forName(providerName, true, Launch.classLoader);
            IEnvironmentTokenProvider provider = (IEnvironmentTokenProvider)th.newInstance();
            this.registerTokenProvider(provider);
         } catch (Throwable var4) {
            logger.error("Error instantiating " + providerName, var4);
         }
      }

      return this;
   }

   public MixinEnvironment registerTokenProvider(IEnvironmentTokenProvider provider) {
      if(provider != null && !this.tokenProviderClasses.contains(provider.getClass().getName())) {
         String providerName = provider.getClass().getName();
         MixinEnvironment.TokenProviderWrapper wrapper = new MixinEnvironment.TokenProviderWrapper(provider, this);
         logger.info("Adding new token provider {} to {}", new Object[]{providerName, this});
         this.tokenProviders.add(wrapper);
         this.tokenProviderClasses.add(providerName);
         Collections.sort(this.tokenProviders);
      }

      return this;
   }

   public Integer getToken(String token) {
      token = token.toUpperCase();
      Iterator i$ = this.tokenProviders.iterator();

      Integer value;
      do {
         if(!i$.hasNext()) {
            return (Integer)this.internalTokens.get(token);
         }

         MixinEnvironment.TokenProviderWrapper provider = (MixinEnvironment.TokenProviderWrapper)i$.next();
         value = provider.getToken(token);
      } while(value == null);

      return value;
   }

   /** @deprecated */
   @Deprecated
   public Set getErrorHandlerClasses() {
      return Mixins.getErrorHandlerClasses();
   }

   public Object getActiveTransformer() {
      return Blackboard.get("mixin.transformer");
   }

   public void setActiveTransformer(IClassTransformer transformer) {
      if(transformer != null) {
         Blackboard.put("mixin.transformer", transformer);
      }

   }

   public MixinEnvironment setSide(MixinEnvironment.Side side) {
      if(side != null && this.getSide() == MixinEnvironment.Side.UNKNOWN && side != MixinEnvironment.Side.UNKNOWN) {
         this.side = side;
      }

      return this;
   }

   public MixinEnvironment.Side getSide() {
      if(this.side == null) {
         MixinEnvironment.Side[] arr$ = MixinEnvironment.Side.values();
         int len$ = arr$.length;

         for(int i$ = 0; i$ < len$; ++i$) {
            MixinEnvironment.Side side = arr$[i$];
            if(side.detect()) {
               this.side = side;
               break;
            }
         }
      }

      return this.side != null?this.side:MixinEnvironment.Side.UNKNOWN;
   }

   public String getVersion() {
      return (String)Blackboard.get("mixin.initialised");
   }

   public boolean getOption(MixinEnvironment.Option option) {
      return this.options[option.ordinal()];
   }

   public void setOption(MixinEnvironment.Option option, boolean value) {
      this.options[option.ordinal()] = value;
   }

   public String getOptionValue(MixinEnvironment.Option option) {
      return option.getStringValue();
   }

   public void setObfuscationContext(String context) {
      this.obfuscationContext = context;
   }

   public String getObfuscationContext() {
      return this.obfuscationContext;
   }

   public String getRefmapObfuscationContext() {
      String overrideObfuscationType = MixinEnvironment.Option.OBFUSCATION_TYPE.getStringValue();
      return overrideObfuscationType != null?overrideObfuscationType:this.obfuscationContext;
   }

   public RemapperChain getRemappers() {
      return this.remappers;
   }

   public void audit() {
      Object activeTransformer = this.getActiveTransformer();
      if(activeTransformer instanceof MixinTransformer) {
         MixinTransformer transformer = (MixinTransformer)activeTransformer;
         transformer.audit();
      }

   }

   public List getTransformers() {
      if(this.transformers == null) {
         this.buildTransformerDelegationList();
      }

      return Collections.unmodifiableList(this.transformers);
   }

   public void addTransformerExclusion(String name) {
      excludeTransformers.add(name);
      this.transformers = null;
   }

   public String unmap(String className) {
      if(this.transformers == null) {
         this.buildTransformerDelegationList();
      }

      return this.nameTransformer != null?this.nameTransformer.unmapClassName(className):className;
   }

   private void buildTransformerDelegationList() {
      logger.debug("Rebuilding transformer delegation list:");
      this.transformers = new ArrayList();
      Iterator i$ = Launch.classLoader.getTransformers().iterator();

      IClassTransformer transformer;
      while(i$.hasNext()) {
         transformer = (IClassTransformer)i$.next();
         String transformerName = transformer.getClass().getName();
         boolean include = true;
         Iterator ignoreTransformer = excludeTransformers.iterator();

         while(true) {
            if(ignoreTransformer.hasNext()) {
               String excludeClass = (String)ignoreTransformer.next();
               if(!transformerName.contains(excludeClass)) {
                  continue;
               }

               include = false;
            }

            boolean ignoreTransformer1 = transformer.getClass().getAnnotation(Resource.class) != null;
            if(include && !ignoreTransformer1 && !transformerName.contains(MixinTransformer.class.getName())) {
               logger.debug("  Adding:    {}", new Object[]{transformerName});
               this.transformers.add(transformer);
               break;
            }

            logger.debug("  Excluding: {}", new Object[]{transformerName});
            break;
         }
      }

      logger.debug("Transformer delegation list created with {} entries", new Object[]{Integer.valueOf(this.transformers.size())});
      i$ = Launch.classLoader.getTransformers().iterator();

      while(i$.hasNext()) {
         transformer = (IClassTransformer)i$.next();
         if(transformer instanceof IClassNameTransformer) {
            logger.debug("Found name transformer: {}", new Object[]{transformer.getClass().getName()});
            this.nameTransformer = (IClassNameTransformer)transformer;
         }
      }

   }

   public String toString() {
      return String.format("%s[%s]", new Object[]{this.getClass().getSimpleName(), this.phase});
   }

   private static MixinEnvironment.Phase getCurrentPhase() {
      if(currentPhase == MixinEnvironment.Phase.NOT_INITIALISED) {
         init(MixinEnvironment.Phase.PREINIT);
      }

      return currentPhase;
   }

   public static void init(MixinEnvironment.Phase phase) {
      if(currentPhase == MixinEnvironment.Phase.NOT_INITIALISED) {
         currentPhase = phase;
         getEnvironment(phase);
         new MixinEnvironment.MixinLogger();
      }

   }

   public static MixinEnvironment getEnvironment(MixinEnvironment.Phase phase) {
      return phase == null?MixinEnvironment.Phase.DEFAULT.getEnvironment():phase.getEnvironment();
   }

   public static MixinEnvironment getDefaultEnvironment() {
      return getEnvironment(MixinEnvironment.Phase.DEFAULT);
   }

   public static MixinEnvironment getCurrentEnvironment() {
      if(currentEnvironment == null) {
         currentEnvironment = getEnvironment(getCurrentPhase());
      }

      return currentEnvironment;
   }

   public static MixinEnvironment.CompatibilityLevel getCompatibilityLevel() {
      return compatibility;
   }

   /** @deprecated */
   @Deprecated
   public static void setCompatibilityLevel(MixinEnvironment.CompatibilityLevel level) throws IllegalArgumentException {
      StackTraceElement[] stackTrace = Thread.currentThread().getStackTrace();
      if(!"org.spongepowered.asm.mixin.transformer.MixinConfig".equals(stackTrace[2].getClassName())) {
         logger.warn("MixinEnvironment::setCompatibilityLevel is deprecated and will be removed. Set level via config instead!");
      }

      if(level != compatibility && level.isAtLeast(compatibility)) {
         if(!level.isSupported()) {
            throw new IllegalArgumentException("The requested compatibility level " + level + " could not be set. Level is not supported");
         }

         compatibility = level;
         logger.info("Compatibility level set to {}", new Object[]{level});
      }

   }

   static void gotoPhase(MixinEnvironment.Phase phase) {
      if(phase != null && phase.ordinal >= 0) {
         if(phase.ordinal > getCurrentPhase().ordinal) {
            MixinBootstrap.addProxy();
         }

         if(phase == MixinEnvironment.Phase.DEFAULT) {
            Logger log = (Logger)LogManager.getLogger("FML");
            log.removeAppender(MixinEnvironment.MixinLogger.appender);
         }

         currentPhase = phase;
         currentEnvironment = getEnvironment(getCurrentPhase());
      } else {
         throw new IllegalArgumentException("Cannot go to the specified phase, phase is null or invalid");
      }
   }


   // $FF: synthetic class
   static class SyntheticClass_1 {
   }

   static class MixinLogger {

      static MixinEnvironment.MixinLogger.MixinAppender appender = new MixinEnvironment.MixinLogger.MixinAppender("MixinLogger", (Filter)null, (Layout)null);


      public MixinLogger() {
         Logger log = (Logger)LogManager.getLogger("FML");
         appender.start();
         log.addAppender(appender);
      }


      static class MixinAppender extends AbstractAppender {

         protected MixinAppender(String name, Filter filter, Layout layout) {
            super(name, filter, layout);
         }

         public void append(LogEvent event) {
            if(event.getLevel() == Level.DEBUG && "Validating minecraft".equals(event.getMessage().getFormat())) {
               MixinEnvironment.gotoPhase(MixinEnvironment.Phase.INIT);
            }

         }
      }
   }

   static class TokenProviderWrapper implements Comparable {

      private static int nextOrder = 0;
      private final int priority;
      private final int order;
      private final IEnvironmentTokenProvider provider;
      private final MixinEnvironment environment;


      public TokenProviderWrapper(IEnvironmentTokenProvider provider, MixinEnvironment environment) {
         this.provider = provider;
         this.environment = environment;
         ++nextOrder;
         this.order = nextOrder;
         this.priority = provider.getPriority();
      }

      public int compareTo(MixinEnvironment.TokenProviderWrapper other) {
         return other == null?0:(other.priority == this.priority?other.order - this.order:other.priority - this.priority);
      }

      public IEnvironmentTokenProvider getProvider() {
         return this.provider;
      }

      Integer getToken(String token) {
         return this.provider.getToken(token, this.environment);
      }

   }

   public static class EnvironmentStateTweaker implements ITweaker {

      public void acceptOptions(List args, File gameDir, File assetsDir, String profile) {
      }

      public void injectIntoClassLoader(LaunchClassLoader classLoader) {
         MixinBootstrap.getPlatform().injectIntoClassLoader(classLoader);
      }

      public String getLaunchTarget() {
         return "";
      }

      public String[] getLaunchArguments() {
         MixinEnvironment.gotoPhase(MixinEnvironment.Phase.DEFAULT);
         return new String[0];
      }
   }

   public static enum CompatibilityLevel {

      JAVA_6(6, 50, false),
      JAVA_7(7, 51, false) {
         boolean isSupported() {
            return JavaVersion.current() >= 1.7D;
         }
      },
      JAVA_8(8, 52, true) {
         boolean isSupported() {
            return JavaVersion.current() >= 1.8D;
         }
      };
      private final int ver;
      private final int classVersion;
      private final boolean supportsMethodsInInterfaces;
      private MixinEnvironment.CompatibilityLevel maxCompatibleLevel;
      // $FF: synthetic field
      private static final MixinEnvironment.CompatibilityLevel[] $VALUES = new MixinEnvironment.CompatibilityLevel[]{JAVA_6, JAVA_7, JAVA_8};


      private CompatibilityLevel(int ver, int classVersion, boolean resolveMethodsInInterfaces) {
         this.ver = ver;
         this.classVersion = classVersion;
         this.supportsMethodsInInterfaces = resolveMethodsInInterfaces;
      }

      private void setMaxCompatibleLevel(MixinEnvironment.CompatibilityLevel maxCompatibleLevel) {
         this.maxCompatibleLevel = maxCompatibleLevel;
      }

      boolean isSupported() {
         return true;
      }

      public int classVersion() {
         return this.classVersion;
      }

      public boolean supportsMethodsInInterfaces() {
         return this.supportsMethodsInInterfaces;
      }

      public boolean isAtLeast(MixinEnvironment.CompatibilityLevel level) {
         return this.ver >= level.ver;
      }

      public boolean canElevateTo(MixinEnvironment.CompatibilityLevel level) {
         return level != null && this.maxCompatibleLevel != null?level.ver <= this.maxCompatibleLevel.ver:true;
      }

      public boolean canSupport(MixinEnvironment.CompatibilityLevel level) {
         return level == null?true:level.canElevateTo(this);
      }

      // $FF: synthetic method
      CompatibilityLevel(int x2, int x3, boolean x4, MixinEnvironment.SyntheticClass_1 x5) {
         this(x2, x3, x4);
      }

   }

   public static enum Option {

      DEBUG_ALL("debug"),
      DEBUG_EXPORT(DEBUG_ALL, "export"),
      DEBUG_EXPORT_FILTER(DEBUG_EXPORT, "filter", false),
      DEBUG_EXPORT_DECOMPILE(DEBUG_EXPORT, "decompile") {
         boolean getBooleanValue() {
            return Booleans.parseBoolean(System.getProperty(this.property), super.getBooleanValue());
         }
      },
      DEBUG_EXPORT_DECOMPILE_THREADED(DEBUG_EXPORT_DECOMPILE, "async") {
         boolean getBooleanValue() {
            return Booleans.parseBoolean(System.getProperty(this.property), super.getBooleanValue());
         }
      },
      DEBUG_VERIFY(DEBUG_ALL, "verify"),
      DEBUG_VERBOSE(DEBUG_ALL, "verbose"),
      DEBUG_INJECTORS(DEBUG_ALL, "countInjections"),
      DEBUG_STRICT(DEBUG_ALL, "strict") {
         boolean getBooleanValue() {
            return this.getLocalBooleanValue();
         }
      },
      DEBUG_UNIQUE(DEBUG_STRICT, "unique"),
      DEBUG_TARGETS(DEBUG_STRICT, "targets"),
      DUMP_TARGET_ON_FAILURE("dumpTargetOnFailure"),
      CHECK_ALL("checks"),
      CHECK_IMPLEMENTS(CHECK_ALL, "interfaces"),
      CHECK_IMPLEMENTS_STRICT(CHECK_IMPLEMENTS, "strict") {
         boolean getBooleanValue() {
            return Booleans.parseBoolean(System.getProperty(this.property), super.getBooleanValue());
         }
      },
      IGNORE_CONSTRAINTS("ignoreConstraints"),
      HOT_SWAP("hotSwap"),
      ENVIRONMENT("env") {
         boolean getBooleanValue() {
            return false;
         }
      },
      OBFUSCATION_TYPE(ENVIRONMENT, "obf"),
      DISABLE_REFMAP(ENVIRONMENT, "disableRefMap"),
      DEFAULT_COMPATIBILITY_LEVEL(ENVIRONMENT, "compatLevel"),
      INITIALISER_INJECTION_MODE("initialiserInjectionMode", "default");
      private static final String PREFIX = "mixin";
      final MixinEnvironment.Option parent;
      final String property;
      final String defaultValue;
      final boolean flag;
      final int depth;
      // $FF: synthetic field
      private static final MixinEnvironment.Option[] $VALUES = new MixinEnvironment.Option[]{DEBUG_ALL, DEBUG_EXPORT, DEBUG_EXPORT_FILTER, DEBUG_EXPORT_DECOMPILE, DEBUG_EXPORT_DECOMPILE_THREADED, DEBUG_VERIFY, DEBUG_VERBOSE, DEBUG_INJECTORS, DEBUG_STRICT, DEBUG_UNIQUE, DEBUG_TARGETS, DUMP_TARGET_ON_FAILURE, CHECK_ALL, CHECK_IMPLEMENTS, CHECK_IMPLEMENTS_STRICT, IGNORE_CONSTRAINTS, HOT_SWAP, ENVIRONMENT, OBFUSCATION_TYPE, DISABLE_REFMAP, DEFAULT_COMPATIBILITY_LEVEL, INITIALISER_INJECTION_MODE};


      private Option(String property) {
         this((MixinEnvironment.Option)null, property, true);
      }

      private Option(String property, boolean flag) {
         this((MixinEnvironment.Option)null, property, flag);
      }

      private Option(String property, String defaultStringValue) {
         this((MixinEnvironment.Option)null, property, false, defaultStringValue);
      }

      private Option(MixinEnvironment.Option parent, String property) {
         this(parent, property, true);
      }

      private Option(MixinEnvironment.Option parent, String property, boolean flag) {
         this(parent, property, flag, (String)null);
      }

      private Option(MixinEnvironment.Option parent, String property, String defaultStringValue) {
         this(parent, property, false, defaultStringValue);
      }

      private Option(MixinEnvironment.Option parent, String property, boolean flag, String defaultStringValue) {
         this.parent = parent;
         this.property = (parent != null?parent.property:"mixin") + "." + property;
         this.defaultValue = defaultStringValue;
         this.flag = flag;

         int depth;
         for(depth = 0; parent != null; ++depth) {
            parent = parent.parent;
         }

         this.depth = depth;
      }

      MixinEnvironment.Option getParent() {
         return this.parent;
      }

      String getProperty() {
         return this.property;
      }

      public String toString() {
         return this.flag?String.valueOf(this.getBooleanValue()):this.getStringValue();
      }

      protected boolean getLocalBooleanValue() {
         return Booleans.parseBoolean(System.getProperty(this.property), false);
      }

      protected boolean getInheritedBooleanValue() {
         return this.parent != null && this.parent.getBooleanValue();
      }

      boolean getBooleanValue() {
         return this.getLocalBooleanValue() || this.getInheritedBooleanValue();
      }

      String getStringValue() {
         return this.parent != null && !this.parent.getBooleanValue()?this.defaultValue:System.getProperty(this.property, this.defaultValue);
      }

      Enum getEnumValue(Enum defaultValue) {
         String value = System.getProperty(this.property, defaultValue.name());

         try {
            return Enum.valueOf(defaultValue.getClass(), value.toUpperCase());
         } catch (IllegalArgumentException var4) {
            return defaultValue;
         }
      }

      // $FF: synthetic method
      Option(MixinEnvironment.Option x2, String x3, MixinEnvironment.SyntheticClass_1 x4) {
         this((MixinEnvironment.Option)x2, (String)x3);
      }

      // $FF: synthetic method
      Option(String x2, MixinEnvironment.SyntheticClass_1 x3) {
         this(x2);
      }

   }

   public static enum Side {

      UNKNOWN {
         protected boolean detect() {
            return false;
         }
      },
      CLIENT {
         protected boolean detect() {
            String sideName = this.getSideName();
            return "CLIENT".equals(sideName);
         }
      },
      SERVER {
         protected boolean detect() {
            String sideName = this.getSideName();
            return "SERVER".equals(sideName) || "DEDICATEDSERVER".equals(sideName);
         }
      };
      // $FF: synthetic field
      private static final MixinEnvironment.Side[] $VALUES = new MixinEnvironment.Side[]{UNKNOWN, CLIENT, SERVER};


      private Side() {
      }

      protected abstract boolean detect();

      protected final String getSideName() {
         Iterator name = ((List)Blackboard.get("Tweaks")).iterator();

         ITweaker tweaker;
         do {
            if(!name.hasNext()) {
               String name1 = this.getSideName("net.minecraftforge.fml.relauncher.FMLLaunchHandler", "side");
               if(name1 != null) {
                  return name1;
               }

               name1 = this.getSideName("cpw.mods.fml.relauncher.FMLLaunchHandler", "side");
               if(name1 != null) {
                  return name1;
               }

               name1 = this.getSideName("com.mumfrey.liteloader.launch.LiteLoaderTweaker", "getEnvironmentType");
               if(name1 != null) {
                  return name1;
               }

               return "UNKNOWN";
            }

            tweaker = (ITweaker)name.next();
            if(tweaker.getClass().getName().endsWith(".common.launcher.FMLServerTweaker")) {
               return "SERVER";
            }
         } while(!tweaker.getClass().getName().endsWith(".common.launcher.FMLTweaker"));

         return "CLIENT";
      }

      private String getSideName(String className, String methodName) {
         try {
            Class ex = Class.forName(className, false, Launch.classLoader);
            Method method = ex.getDeclaredMethod(methodName, new Class[0]);
            return ((Enum)method.invoke((Object)null, new Object[0])).name();
         } catch (Exception var5) {
            return null;
         }
      }

      // $FF: synthetic method
      Side(MixinEnvironment.SyntheticClass_1 x2) {
         this();
      }

   }

   public static class Phase {

      static final MixinEnvironment.Phase NOT_INITIALISED = new MixinEnvironment.Phase(-1, "NOT_INITIALISED");
      public static final MixinEnvironment.Phase PREINIT = new MixinEnvironment.Phase(0, "PREINIT");
      public static final MixinEnvironment.Phase INIT = new MixinEnvironment.Phase(1, "INIT");
      public static final MixinEnvironment.Phase DEFAULT = new MixinEnvironment.Phase(2, "DEFAULT");
      static final List phases = ImmutableList.of(PREINIT, INIT, DEFAULT);
      final int ordinal;
      final String name;
      private MixinEnvironment environment;


      private Phase(int ordinal, String name) {
         this.ordinal = ordinal;
         this.name = name;
      }

      public String toString() {
         return this.name;
      }

      public static MixinEnvironment.Phase forName(String name) {
         Iterator i$ = phases.iterator();

         MixinEnvironment.Phase phase;
         do {
            if(!i$.hasNext()) {
               return null;
            }

            phase = (MixinEnvironment.Phase)i$.next();
         } while(!phase.name.equals(name));

         return phase;
      }

      MixinEnvironment getEnvironment() {
         if(this.ordinal < 0) {
            throw new IllegalArgumentException("Cannot access the NOT_INITIALISED environment");
         } else {
            if(this.environment == null) {
               this.environment = new MixinEnvironment(this);
            }

            return this.environment;
         }
      }

   }
}
